package com.cisco.c3m.est;

/*
 * EncodingException.java
 *
 *  Created on: July 1, 2014
 *      Author: foleyj
 *
 * Copyright (c) 2014 by cisco Systems, Inc.
 * All rights reserved.
 *
 */

public class EncodingException extends Exception {
	/**
	 * 
	 */
	private static final long serialVersionUID = -6015972426836329665L;
	
	public EncodingException(String message) {
        super(message);
    }
}
